%inline %{
/* outputarrs.i */
%}

%{
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "compass.h"
#define pyarrSUB(i, j) SUBSCRIPT2(pyarr, i, j, PyObject*)

#define STRENGTH 0
#define ORIENTATION 1
#define ABNORMALITY 2
#define UNCERTAINTY 3
#define DEFAULTCLUSTERS 10

static void
FreeOutputArrays(PyObject** pyout, int num_outputs)
{
    int i;

    for (i = 0; i < num_outputs; i++) {
        PyArray_XDECREF((PyArrayObject*) pyout[i]);
    }
    free(pyout);
}

/* There are 4 output arrays. Each array is a Numeric array of shape
   (numsigmas, numangles). The entries are Numeric array of shape (rows,
   cols, numresponses) where rows, cols, and numresponses can vary from
   array to array. In the original Compass code, this stuff was mostly in
   compassmex.c.
*/
static PyObject**
CreateOutputArrays(int numsigmas, int numangles, int imgrows, int imgcols,
      int num_outputs, double* spacing, int maxradius, int nwedges)
{
    int i, j, k;
    int dims[3];
    PyObject** pyout;
    PyArrayObject* pyarr;

    pyout = (PyObject**) malloc(4 * sizeof(PyObject*));
    if (pyout == NULL) {
        PyErr_NoMemory();
        return NULL;
    }
    /* NULL all the PyObject* entries so I can use Py_XDECREF. */
    for (i=0; i<4; i++)
        pyout[i] = NULL;

    /* Allocate output arguments. */
    for (i = 0; i < num_outputs; i++) {
        /* Create a 2 dim array (numangles, numsigmas) */
        dims[0] = numangles;
        dims[1] = numsigmas;
        pyout[i] = (PyObject*) PyArray_FromDims(2, dims, PyArray_OBJECT);
        if (pyout[i] == NULL) {
            for (j=0; j<i; j++) Py_XDECREF(pyout[j]);
            free(pyout);
            return NULL;
        }
        /* NULL all the PyObject* entries so I can use Py_XDECREF. */
        pyarr = (PyArrayObject*) pyout[i];
        for (j=0; j<pyarr->dimensions[0]; j++)
            for (k=0; k<pyarr->dimensions[1]; k++)
                pyarrSUB(j, k) = NULL;
    }

    /* Something is a bit weird about the sizes. Its not quite like a
       sliding window operator. In compassmex.c, it says that for the
       default case we are using
           dimensions[0] = maxradius
           dimensions[2] = imgrows - maxradius
           dims[0] = (int)(dimensions[2] - dimensions[0])/(int)spacing[j] + 1;
       which becomes
           dims[0] = (imgrows - 2*maxradius)/(int)spacing[j] + 1;
       I will use this & hold my breath.
    */
    for (i = 0; i < num_outputs; i++) {
        pyarr = (PyArrayObject*) pyout[i];
        for (j=0; j<pyarr->dimensions[0]; j++) { /* sigmas */
            for (k=0; k<pyarr->dimensions[1]; k++) { /* angles */
                /* Create a 3 dim array: (rows, cols, responses). */
                dims[0] = (int) (imgrows - 2*maxradius)/(int) spacing[j] + 1;
                dims[1] = (int) (imgcols - 2*maxradius)/(int) spacing[j] + 1;
                dims[2] = (i == STRENGTH || i == ABNORMALITY) ? 1 : MAXRESPONSES;
                pyarrSUB(j, k) =
                    (PyObject*) PyArray_FromDims(3, dims, PyArray_DOUBLE);
                if (pyarrSUB(j, k) == NULL) {
                    FreeOutputArrays(pyout, num_outputs);
                    return NULL;
                }
            }
        }
    }
    return pyout;
}
#undef pyarrSUB
%}
